import pandas as pd
from dataclasses import dataclass
from typing import TypedDict, List, Optional
from athina.interfaces.data import DataPoint


class EvalResultMetric(TypedDict):
    """
    Represents the LLM evaluation result metric.
    """

    id: str
    value: float

class DatapointFieldAnnotation(TypedDict):
    """
    The annotations to be logged for the datapoint field.
    """

    field_name: str
    text: str
    annotation_type: str
    annotation_note: str

class EvalResult(TypedDict):
    """
    Represents the LLM evaluation result.
    """

    name: str
    display_name: str
    data: DataPoint
    failure: Optional[bool]
    reason: str
    runtime: int
    model: Optional[str]
    metrics: List[EvalResultMetric]
    datapoint_field_annotations: Optional[List[DatapointFieldAnnotation]]

@dataclass
class BatchRunResult:
    """
    Represents the result of a batch run of LLM evaluation.
    """

    eval_request_id: str
    eval_results: List[EvalResult]

    def to_df(self):
        """
        Converts the batch run result to a Pandas DataFrame, including data and dynamic metrics.
        """
        pd.set_option('display.max_colwidth', 500)

        df_data = []
        for item in self.eval_results:
            # Start with dynamic fields from the 'data' dictionary
            entry = {key: value for key, value in item['data'].items()}

            # Add fixed fields
            entry.update({
                'display_name': item['display_name'],
                'failed': item['failure'] if 'failure' in item else None,
                'grade_reason': item['reason'],
                'runtime': item['runtime'],
                'model': item['model'] if 'model' in item else None,
            })

            # Add dynamic metrics
            for metric in item['metrics']:
                entry[metric['id']] = metric['value']

            df_data.append(entry)

        df = pd.DataFrame(df_data)
        return df

class EvalPerformanceReport(TypedDict):
    """
    Represents the performance metrics for an evaluation.
    """

    true_positives: int
    false_positives: int
    true_negatives: int
    false_negatives: int
    accuracy: float
    precision: float
    recall: float
    f1_score: float
    runtime: int
    dataset_size: int


from abc import ABC, abstractmethod
from enum import Enum
from typing import TypedDict, List
import json
from athina.interfaces.data import DataPoint


class LoadFormat(Enum):
    """Supported load formats."""

    JSON = "json"
    CSV = "csv"
    DICT = "dict"
    ATHINA = "athina"


class Loader(ABC):
    """Abstract base class for data loaders."""

    @property
    def processed_dataset(self) -> List[DataPoint]:
        """
        Returns the processed dataset.
        """
        return self._processed_dataset

    @property
    def raw_dataset(self):
        """
        Returns the raw dataset.
        """
        return self._raw_dataset

    @abstractmethod
    def process(self) -> List[DataPoint]:
        """Prepare dataset to be consumed by evaluators."""
        pass

    def load(self, format: str, **kwargs) -> List[DataPoint]:
        """
        Loads data based on the format specified.
        """
        if format == LoadFormat.JSON.value:
            return self.load_json(**kwargs)
        elif format == LoadFormat.DICT.value:
            return self.load_dict(**kwargs)
        elif format == LoadFormat.CSV.value:
            return self.load_csv(**kwargs)
        elif format == LoadFormat.ATHINA.value:
            return self.load_athina_inferences(**kwargs)
        else:
            raise NotImplementedError("This file format has not been supported yet.")

    def load_json(self, filename: str) -> List[DataPoint]:
        """
        Loads and processes data from a JSON file.

        Raises:
            FileNotFoundError: If the specified JSON file is not found.
            json.JSONDecodeError: If there's an issue decoding the JSON.
        """
        try:
            with open(filename, "r") as f:
                self._raw_dataset = json.load(f)
                self.process()
                return self._processed_dataset
        except (FileNotFoundError, json.JSONDecodeError) as e:
            print(f"Error loading JSON: {e}")

    def load_dict(self, data: list) -> List[DataPoint]:
        """
        Loads and processes data from a list of dictionaries.
        """
        self._raw_dataset = data
        self.process()
        return self._processed_dataset

    def load_csv(self) -> List[DataPoint]:
        """
        Placeholder for loading data from a CSV file.

        Raises:
            NotImplementedError: This method has not been implemented yet.
        """
        raise NotImplementedError("This method has not been implemented yet.")

    def load_pandas(self) -> List[DataPoint]:
        """
        Placeholder for loading data from a pandas DataFrame.

        Raises:
            NotImplementedError: This method has not been implemented yet.
        """
        raise NotImplementedError("This method has not been implemented yet.")



from typing import List, TypedDict, Optional
from athina.helpers.athina_logging_helper import AthinaLoggingHelper
from athina.evals.llm.llm_evaluator import LlmEvaluator
from athina.interfaces.result import EvalResult, BatchRunResult
from athina.interfaces.data import DataPoint
from athina.interfaces.athina import AthinaExperiment
from athina.services.athina_api_service import AthinaApiService


class DataPointWithEvalResults(TypedDict):
    """A data point with its evaluation results."""

    data_point: DataPoint
    eval_results: List[EvalResult]


class LlmEvaluatorDescription(TypedDict):
    """A description of an LLM evaluator."""

    name: str
    display_name: str


class LlmBatchEvalResult(TypedDict):
    """Result of running a batch of LLM evaluations."""

    results: List[EvalResult]
    total_runtime: float
    passed_evals: int
    failed_evals: int
    total_evals: int
    total_datapoints: int


class EvalRunner:
    @staticmethod
    def batch_eval_result(
        eval_results: List[EvalResult],
    ) -> LlmBatchEvalResult:
        """
        Calculate metrics for a batch of LLM evaluations.

        Args:
            datapoints_with_eval_results: A list of DataPointWithEvalResults objects.

        Returns:
            A LlmBatchEvalResult object.
        """
        total_runtime = 0
        passed_evals = 0
        failed_evals = 0
        total_evals = 0

        # Iterate through each DataPointWithEvalResults
        for eval_result in eval_results:
            total_evals += 1
            total_runtime += eval_result.get("runtime", 0)

            # Counting passed and failed evaluations
            if eval_result.get("failure"):
                failed_evals += 1
            else:
                passed_evals += 1

        total_datapoints = len(eval_result)

        return LlmBatchEvalResult(
            results=eval_result,
            total_runtime=total_runtime,
            passed_evals=passed_evals,
            failed_evals=failed_evals,
            total_evals=total_evals,
            total_datapoints=total_datapoints,
        )

    @staticmethod
    def run_suite(
        evals: List[LlmEvaluator],
        data: List[DataPoint],
        experiment: Optional[AthinaExperiment] = None,
        max_parallel_evals: int = 1,
    ) -> List[LlmBatchEvalResult]:
        """
        Run a suite of LLM evaluations against a dataset.

        Args:
            evals: A list of LlmEvaluator objects.
            data: A list of data points.

        Returns:
            A list of LlmBatchEvalResult objects.
        """
        # Create eval request
        eval_suite_name = "llm_eval_suite" + "_" + ",".join(eval.name for eval in evals)
        eval_request_id = AthinaLoggingHelper.create_eval_request(
            eval_name=eval_suite_name,
            request_data={"data": data},
            request_type="suite",
        )

        # Log experiment
        if experiment is not None:
            AthinaLoggingHelper.log_experiment(
                eval_request_id=eval_request_id,
                experiment=experiment,
            )

        AthinaApiService.log_usage(eval_name=eval_suite_name, run_type="suite")

        batch_results = []
        for eval in evals:
            # Log usage to Athina for analytics

            # Validate the dataset against the required args
            eval._validate_batch_args(data)

            # Run the evaluations
            if max_parallel_evals > 1:
                eval_results = eval._run_batch_generator_async(data, max_parallel_evals)
            else:
                eval_results = list(eval._run_batch_generator(data))

            # Log evaluation results to Athina
            AthinaLoggingHelper.log_eval_results(
                eval_request_id=eval_request_id,
                eval_results=eval_results,
            )

            batch_results.append(BatchRunResult(
                eval_request_id=eval_request_id,
                eval_results=eval_results,
            ))

        return EvalRunner.batch_eval_result(eval_results=eval_results)



from .rag_loader import RagLoader
from .ragas_loader import RagasLoader
from .response_loader import ResponseLoader
from .summary_loader import SummaryLoader
from .loader import Loader, LoadFormat

__all__ = [
    "RagLoader",
    "RagasLoader",
    "ResponseLoader",
    "SummaryLoader",
    "Loader",
    "LoadFormat",
]



import yaml

CONFIG_FILE_NAME = "athina_config.yml"


class ConfigHelper:
    @staticmethod
    def load_config():
        try:
            with open(CONFIG_FILE_NAME, "r") as file:
                config = yaml.safe_load(file)

            if config is None:
                config = {}
            return config
        except:
            return {}

    @staticmethod
    def load_config_field(field: str):
        try:
            config = ConfigHelper.load_config()
            return config[field]
        except Exception as e:
            return None

    @staticmethod
    def load_openai_api_key():
        return ConfigHelper.load_config_field("openai_api_key")

    @staticmethod
    def load_athina_api_key():
        return ConfigHelper.load_config_field("athina_api_key")

    @staticmethod
    def load_llm_engine():
        return ConfigHelper.load_config_field("llm_engine")

    @staticmethod
    def save_config(config_data):
        with open(CONFIG_FILE_NAME, "w") as file:
            yaml.dump(config_data, file)

    @staticmethod
    def is_set():
        try:
            with open(CONFIG_FILE_NAME, "r") as file:
                config = yaml.safe_load(file)

            if config is None or config == {}:
                return False
            else:
                return True
        except:
            return False



# athina/evals/__init__.py
from athina.evals.llm.does_response_answer_query.evaluator import (
    DoesResponseAnswerQuery,
)
from athina.evals.llm.context_contains_enough_information.evaluator import (
    ContextContainsEnoughInformation,
)
from athina.evals.base_evaluator import BaseEvaluator
from athina.evals.llm.faithfulness.evaluator import Faithfulness
from athina.evals.llm.grading_criteria.evaluator import GradingCriteria
from athina.evals.llm.custom_prompt.evaluator import CustomPrompt
from athina.evals.llm.summary_accuracy.evaluator import SummaryAccuracy
from athina.evals.llm.groundedness.evaluator  import Groundedness
from athina.evals.ragas.context_relevancy.evaluator import RagasContextRelevancy
from athina.evals.ragas.answer_relevancy.evaluator import RagasAnswerRelevancy
from athina.evals.ragas.context_precision.evaluator import RagasContextPrecision
from athina.evals.ragas.faithfulness.evaluator import RagasFaithfulness
from athina.evals.ragas.context_recall.evaluator import RagasContextRecall
from athina.evals.ragas.answer_semantic_similarity.evaluator import RagasAnswerSemanticSimilarity
from athina.evals.ragas.answer_correctness.evaluator import RagasAnswerCorrectness
from athina.evals.ragas.harmfulness.evaluator import RagasHarmfulness
from athina.evals.ragas.maliciousness.evaluator import RagasMaliciousness
from athina.evals.ragas.coherence.evaluator import RagasCoherence 
from athina.evals.ragas.conciseness.evaluator import RagasConciseness
from athina.evals.function.function_evaluator import FunctionEvaluator
from athina.evals.llm.llm_evaluator import LlmEvaluator
from athina.evals.function.wrapper import ContainsAny, Regex, ContainsAll, Contains, ContainsNone, ContainsJson, ContainsEmail, IsJson, IsEmail, NoInvalidLinks, ContainsLink, ContainsValidLink, Equals, StartsWith, EndsWith, LengthLessThan, LengthGreaterThan, ApiCall

__all__ = [
    "BaseEvaluator",
    "LlmEvaluator",
    "DoesResponseAnswerQuery",
    "SummaryAccuracy",
    "ContextContainsEnoughInformation",
    "Faithfulness",
    "RagasContextRelevancy",
    "RagasAnswerRelevancy",
    "RagasContextPrecision",
    "RagasFaithfulness",
    "RagasContextRecall",
    "RagasAnswerSemanticSimilarity",
    "RagasAnswerCorrectness",
    "RagasHarmfulness",
    "RagasMaliciousness",
    "RagasCoherence",
    "RagasConciseness",
    "FunctionEvaluator",
    "GradingCriteria",
    "Groundedness",
    "CustomPrompt",
    "ContainsAny",
    "Regex",
    "ContainsAll",
    "Contains",
    "ContainsNone",
    "ContainsJson",
    "ContainsEmail",
    "IsJson",
    "IsEmail",
    "NoInvalidLinks",
    "ContainsLink",
    "ContainsValidLink",
    "Equals",
    "StartsWith",
    "EndsWith",
    "LengthLessThan",
    "LengthGreaterThan",
    "ApiCall"
]



#!/usr/bin/env python3

import argparse
from athina.helpers.config import ConfigHelper
from athina.helpers.run_helper import RunHelper
from athina.helpers.kwparser import KeyValueAction
from athina.interfaces.model import Model
from athina.loaders import LoadFormat
from typing import Optional


def main():
    parser = argparse.ArgumentParser(
        prog="athina",
        description="Evaluation framework for your LLM-powered applications",
    )

    subparsers = parser.add_subparsers(title="commands", dest="command")

    # athina init
    parser_init = subparsers.add_parser("init", help="Configure settings")
    parser_init.set_defaults(func=init)

    # athina config
    parser_config = subparsers.add_parser("config", help="Configure settings")
    parser_config.set_defaults(func=config)

    # athina list
    parser_config = subparsers.add_parser("list", help="Lists all available evals")
    parser_config.set_defaults(func=list)

    # athina run [eval_name] [kwargs]
    parser_run = subparsers.add_parser("run", help="Run an eval suite")

    # Add the 'eval_name' positional argument
    parser_run.add_argument(
        "eval_name",
        type=str,
        help="The name of the eval or eval suite to run",
    )

    # Add the 'kwargs' argument for key=value pairs
    parser_run.add_argument(
        "kwargs",
        nargs="*",
        action=KeyValueAction,
        help="Additional named arguments as key=value pairs",
    )

    # Add the '--format' optional argument
    parser_run.add_argument(
        "--model",
        type=str,
        choices=[
            Model.GPT35_TURBO.value,
            Model.GPT4.value,
            Model.GPT4_1106_PREVIEW.value,
        ],
        help="LLM model for evaluation",
    )

    # Add the '--format' optional argument
    parser_run.add_argument(
        "--format",
        type=str,
        choices=[
            LoadFormat.JSON.value,
            LoadFormat.CSV.value,
            LoadFormat.ATHINA.value,
        ],
        help="Output format type",
    )

    # Add the '--filename' optional argument
    parser_run.add_argument(
        "--filename",
        type=str,
        help="Path to the file",
    )

    # Set the default function to be called
    parser_run.set_defaults(func=run_delegator)

    # Parse the arguments
    args = parser.parse_args()

    if hasattr(args, "func"):
        args.func(args)
    else:
        parser.print_help()


def init(args):
    """Initializes Athina and sets the necessary configuration variables"""
    config_data = ConfigHelper.load_config()

    openai_api_key = input("Enter your OpenAI API key: ")
    config_data["openai_api_key"] = openai_api_key

    athina_api_key = input("Enter your Athina API key: ")
    config_data["athina_api_key"] = athina_api_key

    config_data["llm_engine"] = "gpt-4-1106-preview"

    # Add other configuration prompts as needed

    ConfigHelper.save_config(config_data)
    print("Configuration updated successfully. See athina_config.yml for details.")


def config(args):
    """Prints the current configuration"""
    config_data = ConfigHelper.load_config()
    print(config_data)


def list(args):
    """Lists all available evals"""
    evals = RunHelper.all_evals()
    evals_list = "- "
    evals_list += "\n- ".join(evals)
    print(evals_list)


def run_delegator(args):
    """Delegates the run command to the appropriate function"""

    if not ConfigHelper.is_set():
        print("Please run 'athina init' to configure your API keys")
        return

    # Load the eval model
    model = ConfigHelper.load_llm_engine()
    if args.model is not None:
        model = args.model

    filename = args.filename if args.filename else None

    # Check if format is 'athina'
    if args.format == "athina":
        run_batch(args.eval_name, model, format="athina")
        return

    # Check if both format and filename are set
    elif args.format is not None and filename is not None:
        run_batch(args.eval_name, model, format=args.format, filename=filename)
        return

    # If format and filename are both None, call run_datapoint with kwargs
    elif args.format is None and filename is None:
        run_datapoint(args.eval_name, model, **dict(args.kwargs))
        return

    elif args.format is not None and filename is None:
        raise Exception("Filename must be specified for batch process")
        return

    else:
        raise Exception("Invalid run args")


# Define the run_batch function
def run_batch(
    eval_name: str, model: str, format: str, filename: Optional[str] = None, **kwargs
):
    # Implementation for running batch process
    try:
        print(
            f"Running batch with format={format}, model={model}, filename={filename}, kwargs={kwargs}"
        )

        RunHelper.run_eval_on_batch(
            eval_name=eval_name,
            model=model,
            format=format,
            filename=filename,
            **kwargs
        )
    except Exception as e:
        print(f"{e}")
        return


def run_datapoint(eval_name: str, model: str, **kwargs):
    """Runs a single eval on a single datapoint"""
    try:
        print(f"Running single with {eval_name} and kwargs {kwargs}")
        RunHelper.run_eval(eval_name, model, kwargs)
    except Exception as e:
        print(f"{e}")
        return


if __name__ == "__main__":
    main()



from typing import List, Tuple, Optional
from athina.llms.openai_service import OpenAiService
from athina.llms.abstract_llm_service import AbstractLlmService
from .question_answerer import QuestionAnswerer

class QuestionAnswererBulk(QuestionAnswerer):

    _llm_service: AbstractLlmService

    """
    This class responds to a list of closed-ended (Y/N) questions based on a provided context.
    It does so using a single LLM inference call, and retrieving a JSON dictionary of all responses.
    """

    # Pre-defined prompts for OpenAI's GPT model
    SYSTEM_MESSAGE = """ 
        You are an expert at responding to closed-ended (Yes/No) questions using ONLY the provided context.
    """

    USER_MESSAGE_TEMPLATE = """
        Let's think step by step.
        1. Consider the following: 
           Questions: {}.
           Context: {}.
        2. Respond to each question from the provided 'questions', using either 
           'Yes', 'No', or 'Unknown', based ONLY on the given context.
        3. Return a JSON object in the following format: 
            [question1]: answer1,
            [question2]: answer2,
            ...
    """

    def __init__(
        self, 
        model: str = "gpt-4-1106-preview",
        llm_service: Optional[AbstractLlmService] = None
    ):
        """
        Initialize the QuestionAnswerer class.
        """
        self._model = model
        if llm_service is None:
            self._llm_service = OpenAiService()
        else:
            self._llm_service = llm_service

    def answer(self, questions: List[str], context: str) -> Tuple[dict, dict]:
        """
        Respond to each question from the provided 'questions' given the context.
        """

        questions_str = '\n'.join(questions)
        user_message = self.USER_MESSAGE_TEMPLATE.format(questions_str, context)
        messages = [
            {"role": "system", "content": self.SYSTEM_MESSAGE},
            {"role": "user", "content": user_message},
        ]

        # Extract JSON object from LLM response
        json_response = self._llm_service.json_completion(
            model=self._model,
            messages=messages,
        )

        if json_response is None:
            raise Exception("No response from LLM")

        output = {}
        simple_output = {}
        for i in range(len(questions)):
            question = questions[i]
            try:
                answer = json_response[question]
                output[question] = {
                    "answer": answer,
                    "explanation": None
                }
                simple_output[question] = answer
            except:
                output[question] = {
                    "answer": "Error",
                    "explanation": None,
                }
                simple_output[question] = "Error"

        return output, simple_output



from typing import List, Optional, TypedDict
from athina.interfaces.athina import AthinaFilters
from athina.interfaces.data import DataPoint
from .loader import Loader


class RagasDataPoint(TypedDict):
    """Data point for a single RAG inference."""

    query: Optional[str]
    contexts: Optional[List[str]]
    response: Optional[str]
    expected_response = Optional[str]


class RagasLoader(Loader):
    """
    This class is a data loader for retrieval augmented generation (RAG) datasets.

    Attributes:
        col_query (str): The column name corresponding to the user's query.
        col_contexts (List[str]): The column name corresponding to the retrieved contexts.
        col_response (str): The column name corresponding to the response.
        col_expected_response (str): The column name corresponding to the expected response.
        raw_dataset (dict): The raw dataset as loaded from the source.
        processed_dataset (list): The processed dataset with queries, context, response and other attributes if present.
    """

    def __init__(
        self,
        col_query="query",
        col_contexts="contexts",
        col_response="response",
        col_expected_response="expected_response"
    ):
        """
        Initializes the loader with specified or default column names.
        """
        self.col_query = col_query
        self.col_contexts = col_contexts
        self.col_response = col_response
        self.col_expected_response = col_expected_response
        self._raw_dataset = {}
        self._processed_dataset: List[RagasDataPoint] = []

    def process(self) -> None:
        """
        Transforms the raw data into a structured format. Processes each entry from the raw dataset, and extracts attributes.

        Raises:
            TypeError: If any key is not of expected type.
        """
        for raw_instance in self._raw_dataset:
            # Check for the type of columns in raw_instance
            if self.col_query in raw_instance and not isinstance(raw_instance.get(self.col_query), str):
                raise TypeError(f"'{self.col_query}' is not of type string.")
            if self.col_contexts in raw_instance:
                if not isinstance(raw_instance.get(self.col_contexts), list):
                    raise TypeError(f"'{self.col_contexts}' is not of type list.")
                if not all(isinstance(element, str) for element in raw_instance.get(self.col_contexts)):
                    raise TypeError(f"Not all elements in '{self.col_context}' are of type string.")
            if self.col_response in raw_instance and not isinstance(raw_instance.get(self.col_response), str):
                raise TypeError(f"'{self.col_response}' is not of type string.")
            if self.col_expected_response in raw_instance and not isinstance(raw_instance.get(self.col_expected_response), str):
                raise TypeError(f"'{self.col_expected_response}' is not of type string.")
            

            # Create a processed instance
            processed_instance = {
                "query": raw_instance.get(self.col_query, None),
                "contexts": raw_instance.get(self.col_contexts, None),
                "response": raw_instance.get(self.col_response, None),
                "expected_response": raw_instance.get(self.col_expected_response, None)
            }


            # Store the results
            self._processed_dataset.append(processed_instance)

    def load_athina_inferences(
        self,
        filters: Optional[AthinaFilters] = None,
        limit: int = 10,
        context_key: Optional[str] = None,
    ):
        """
        Load data from Athina API.
        By default, this will fetch the last 10 inferences from the API.
        """
        pass